/*
 * Copyright (C) 2012 Charlie Curtsinger
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef stabilizer2_Metadata_h
#define stabilizer2_Metadata_h

#include "Heaps.h"

#include <memory>
#include <limits>

// Class that all stabilizer metadata types must inherit from
class Metadata {
public:
	inline static void* operator new(size_t sz) {
		return MD_malloc(sz);
	}
	
	inline static void operator delete(void *p) {
		MD_free(p);
	}
};

// STL Allocator for the metadata heap
template<class T>
class MDAllocator {
public:
	// type definitions
	typedef T        value_type;
	typedef T*       pointer;
	typedef const T* const_pointer;
	typedef T&       reference;
	typedef const T& const_reference;
	typedef std::size_t    size_type;
	typedef std::ptrdiff_t difference_type;
	
	// rebind allocator to type U
	template <class U>
	struct rebind {
		typedef MDAllocator<U> other;
	};
	
	// return address of values
	pointer address (reference value) const {
		return &value;
	}
	const_pointer address (const_reference value) const {
		return &value;
	}
	
	MDAllocator() throw() {}
	
	MDAllocator(const MDAllocator&) throw() {}
	
	template <class U>
	MDAllocator (const MDAllocator<U>&) throw() {}
	
	~MDAllocator() throw() {}
	
	// return maximum number of elements that can be allocated
	size_type max_size () const throw() {
		return std::numeric_limits<std::size_t>::max() / sizeof(T);
	}
	
	// allocate but don't initialize num elements of type T
	pointer allocate (size_type num, const void* = 0) {
		return (pointer)MD_malloc(num*sizeof(T));
	}
	
	// initialize elements of allocated storage p with value value
	void construct (pointer p, const T& value) {
		// initialize memory with placement new
		new((void*)p)T(value);
	}
	
	// destroy elements of initialized storage p
	void destroy (pointer p) {
		// destroy objects by calling their destructor
		p->~T();
	}
	
	// deallocate storage p of deleted elements
	void deallocate (pointer p, size_type num) {
		MD_free((void*)p);
	}
};

#endif
